/*	$OpenBSD: database.c,v 1.16 2004/06/22 03:15:33 avsm Exp $	*/

/* Copyright 1988,1990,1993,1994 by Paul Vixie
 * All rights reserved
 */

/*
 * Copyright (c) 2004 by Internet Systems Consortium, Inc. ("ISC")
 * Copyright (c) 1997,2000 by Internet Software Consortium, Inc.
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#if !defined(lint) && !defined(LINT)
static char const rcsid[] = "$OpenBSD: database.c,v 1.16 2004/06/22 03:15:33 avsm Exp $";
#endif

/* vix 26jan87 [RCS has the log]
 */

#include "cron.h"

#define HASH(a,b) ((a)+(b))

struct spooldir {
	char *path;
	char *uname;
	char *fname;
};

static struct spooldir spools[] = {
	{SPOOL_DIR, NULL, NULL},
	{"/etc/cron.d", "root", SYSUSERNAME},
	{NULL, NULL, NULL}
};

static	void		process_crontab(const char *, const char *,
					const char *, struct stat *,
					cron_db *, cron_db *);

void
load_database(cron_db *old_db) {
	struct stat statbuf, syscron_stat;
	cron_db new_db;
	DIR_T *dp;
	DIR *dir;
	user *u, *nu;

	struct spooldir *p = spools;
	time_t maxtime;

	Debug(DLOAD, ("[%ld] load_database()\n", (long)getpid()))

	/* track system crontab file
	 */
	if (stat(SYSCRONTAB, &syscron_stat) < OK)
		syscron_stat.st_mtime = 0;

	maxtime = syscron_stat.st_mtime;
	while (p->path) {
		if (stat(p->path, &statbuf) < OK) {
			log_it("CRON", getpid(), "STAT FAILED", p->path);
			(void) exit(ERROR_EXIT);
		}
		if (statbuf.st_mtime > maxtime)
			maxtime = statbuf.st_mtime;
		p++;
	}

	/* if spooldir's mtime has not changed, we don't need to fiddle with
	 * the database.
	 *
	 * Note that old_db->mtime is initialized to 0 in main(), and
	 * so is guaranteed to be different than the stat() mtime the first
	 * time this function is called.
	 */
	if (old_db->mtime == maxtime) {
		Debug(DLOAD, ("[%ld] spool dir mtime unch, no load needed.\n",
			      (long)getpid()))
		return;
	}

	/* something's different.  make a new database, moving unchanged
	 * elements from the old database, reloading elements that have
	 * actually changed.  Whatever is left in the old database when
	 * we're done is chaff -- crontabs that disappeared.
	 */
	new_db.mtime = maxtime;
	new_db.head = new_db.tail = NULL;

	if (syscron_stat.st_mtime) {
		process_crontab(ROOT_USER, SYSUSERNAME, SYSCRONTAB, &syscron_stat,
				&new_db, old_db);
	}

	p = spools;
 	while (p->path) {
 		if (!(dir = opendir(p->path))) {
 			log_it("CRON", getpid(), "OPENDIR FAILED", p->path);
 			(void) exit(ERROR_EXIT);
 		}
 
 		while (NULL != (dp = readdir(dir))) {
 			char	fname[MAXNAMLEN+1], tabname[MAXPATHLEN];
 
 			/* Don't try to parse any files containing a dot ('.')
 			 * or ending with a tilde ('~'). This catches the case
			 * of '.' and '..', as well as preventing the parsing
			 * of many editor files, temporary files and those
			 * saved by RPM package upgrades.
 			 */
 			if ( !dp->d_name[0] /* Shouldn't happen */ || 
 			      strchr(dp->d_name, '.') ||
 			      dp->d_name[strlen(dp->d_name)-1] == '~')
 				continue;

			if (strlcpy(fname, dp->d_name, sizeof fname) >= sizeof fname)
				continue;	/* XXX log? */
 
			if (snprintf(tabname, sizeof tabname, "%s/%s", p->path, fname) >=
				sizeof(tabname))
				continue;	/* XXX log? */

 			process_crontab(p->uname ? p->uname : fname,
 					p->fname ? p->fname : fname,
 					tabname,
 					&statbuf, &new_db, old_db);
 		}
 		closedir(dir);
 		/* we used to keep this dir open all the time, for the sake of
 		 * efficiency.  however, we need to close it in every fork, and
 		 * we fork a lot more often than the mtime of the dir changes.
  		 */
  
 		p++;
  	}
	
	/* if we don't do this, then when our children eventually call
	 * getpwnam() in do_command.c's child_process to verify MAILTO=,
	 * they will screw us up (and v-v).
	 */
	endpwent();

	/* whatever's left in the old database is now junk.
	 */
	Debug(DLOAD, ("unlinking old database:\n"))
	for (u = old_db->head;  u != NULL;  u = nu) {
		Debug(DLOAD, ("\t%s\n", u->name))
		nu = u->next;
		unlink_user(old_db, u);
		free_user(u);
	}

	/* overwrite the database control block with the new one.
	 */
	*old_db = new_db;
	Debug(DLOAD, ("load_database is done\n"))
}

void
link_user(cron_db *db, user *u) {
	if (db->head == NULL)
		db->head = u;
	if (db->tail)
		db->tail->next = u;
	u->prev = db->tail;
	u->next = NULL;
	db->tail = u;
}

void
unlink_user(cron_db *db, user *u) {
	if (u->prev == NULL)
		db->head = u->next;
	else
		u->prev->next = u->next;

	if (u->next == NULL)
		db->tail = u->prev;
	else
		u->next->prev = u->prev;
}

user *
find_user(cron_db *db, const char *name) {
	user *u;

	for (u = db->head;  u != NULL;  u = u->next)
		if (strcmp(u->name, name) == 0)
			break;
	return (u);
}

static void
process_crontab(const char *uname, const char *fname, const char *tabname,
		struct stat *statbuf, cron_db *new_db, cron_db *old_db)
{
	struct passwd *pw = NULL;
	int crontab_fd = OK - 1;
	struct stat lstatbuf;
	user *u;

	if (strcmp(fname, SYSUSERNAME) && !(pw = getpwnam(uname))) {
		/* file doesn't have a user in passwd file.
		 * 		 */
		log_it(fname, getpid(), "ORPHAN", "no passwd entry");
		goto next_crontab;
	}

	if (lstat(tabname, &lstatbuf) < OK) {
		log_it(fname, getpid(), "CAN'T LSTAT", tabname);
		goto next_crontab;
	}
	if (!S_ISREG(lstatbuf.st_mode)) {
		log_it(fname, getpid(), "NOT REGULAR", tabname);
		goto next_crontab;
	}
	if ((!pw && (lstatbuf.st_mode & 07533) != 0400) ||
	    (pw && (lstatbuf.st_mode & 07577) != 0400)) {
		log_it(fname, getpid(), "BAD FILE MODE", tabname);
		goto next_crontab;
	}
	if (pw && lstatbuf.st_nlink != 1) {
		log_it(fname, getpid(), "BAD LINK COUNT", tabname);
		goto next_crontab;
	}

	if ((crontab_fd = open(tabname, O_RDONLY|O_NONBLOCK|O_NOFOLLOW, 0)) < OK) {
		/* crontab not accessible?
		 */
		log_it(fname, getpid(), "CAN'T OPEN", tabname);
		goto next_crontab;
	}

	if (fstat(crontab_fd, statbuf) < OK) {
		log_it(fname, getpid(), "FSTAT FAILED", tabname);
		goto next_crontab;
	}
	if (!S_ISREG(statbuf->st_mode)) {
		log_it(fname, getpid(), "NOT REGULAR", tabname);
		goto next_crontab;
	}
	if ((!pw && (statbuf->st_mode & 07533) != 0400) ||
	    (pw && (statbuf->st_mode & 07577) != 0400)) {
		log_it(fname, getpid(), "BAD FILE MODE", tabname);
		goto next_crontab;
	}
	if (statbuf->st_uid != ROOT_UID && (pw == NULL ||
	    statbuf->st_uid != pw->pw_uid || strcmp(uname, pw->pw_name) != 0)) {
		log_it(fname, getpid(), "WRONG FILE OWNER", tabname);
		goto next_crontab;
	}
	if (pw && statbuf->st_nlink != 1) {
		log_it(fname, getpid(), "BAD LINK COUNT", tabname);
		goto next_crontab;
	}
	if (lstatbuf.st_dev != statbuf->st_dev ||
			lstatbuf.st_ino != statbuf->st_ino) {
		log_it(fname, getpid(), "FILE CHANGED DURING OPEN", tabname);
		goto next_crontab;
	}

	Debug(DLOAD, ("\t%s:", fname))
	u = find_user(old_db, fname);
	if (u != NULL) {
		/* if crontab has not changed since we last read it
		 * in, then we can just use our existing entry.
		 */
		if (u->mtime == statbuf->st_mtime) {
			Debug(DLOAD, (" [no change, using old data]"))
			unlink_user(old_db, u);
			link_user(new_db, u);
			goto next_crontab;
		}

		/* before we fall through to the code that will reload
		 * the user, let's deallocate and unlink the user in
		 * the old database.  This is more a point of memory
		 * efficiency than anything else, since all leftover
		 * users will be deleted from the old database when
		 * we finish with the crontab...
		 */
		Debug(DLOAD, (" [delete old data]"))
		unlink_user(old_db, u);
		free_user(u);
		log_it(fname, getpid(), "RELOAD", tabname);
	}

	u = load_user(crontab_fd, pw, fname, tabname);
	if (u != NULL) {
		u->mtime = statbuf->st_mtime;
		link_user(new_db, u);
	}

 next_crontab:
	if (crontab_fd >= OK) {
		Debug(DLOAD, (" [done]\n"))
		close(crontab_fd);
	}
}
